"""
acl_test
~~~~~~~~~~~~~
Tests the pre-processor that constructs a dictionary of ACL keys from a YAML
file.
"""
from lever import build_acl

import yaml
import unittest


class TestACLContrstruction(unittest.TestCase):
    def test_virtual_removal(self):
        struct = yaml.load("""
        testing:
            virtual: true""")
        acl = build_acl(struct)
        assert 'testing' not in acl

    def test_inherit_virtual(self):
        struct = yaml.load("""
        testing:
            virtual: true
            user:
                key: exists
        testing2:
            inherit: testing
            admin:
                inherit: user""")
        acl = build_acl(struct)
        assert 'testing' not in acl
        assert 'key_exists' in acl['testing2']['admin']

    def test_basic_key(self):
        struct = yaml.load("""
        testing:
            user:
                key: exists""")
        acl = build_acl(struct)
        assert 'key_exists' in acl['testing']['user']

    def test_basic_key_list(self):
        struct = yaml.load("""
        testing:
            user:
                key:
                    - exists""")
        acl = build_acl(struct)
        assert 'key_exists' in acl['testing']['user']

    def test_basic_inherit_role(self):
        struct = yaml.load("""
        testing:
            admin:
                inherit: user
            user:
                key:
                    - exists""")
        acl = build_acl(struct)
        assert 'key_exists' in acl['testing']['admin']

    def test_basic_inherit_list_role(self):
        struct = yaml.load("""
        testing:
            admin:
                inherit:
                    - user
                    - mod
            mod:
                key:
                    - also_exists
            user:
                key:
                    - exists""")
        acl = build_acl(struct)
        assert 'key_exists' in acl['testing']['admin']
        assert 'key_also_exists' in acl['testing']['admin']

    def test_inherit_failure_role(self):
        struct = yaml.load("""
        testing:
            admin:
                inherit:
                    - broken""")
        self.assertRaises(KeyError, build_acl, struct)

    def test_inherit_failure_type(self):
        struct = yaml.load("""
        testing2:
            inherit: doesnt_exist""")
        self.assertRaises(KeyError, build_acl, struct)

    def test_looping_inherit(self):
        struct = yaml.load("""
        testing0:
            inherit: testing3
        testing1:
            inherit: testing0
        testing2:
            inherit: testing1
        testing3:
            inherit: testing2""")
        self.assertRaises(Exception, build_acl, struct)

    def test_key_dict_unsupport(self):
        struct = yaml.load("""
        testing:
            admin:
                key:
                    wont:
                        work""")
        self.assertRaises(Exception, build_acl, struct)
        try:
            build_acl(struct)
        except Exception as e:
            assert 'not supported' in str(e)

    def test_basic_inherit_type(self):
        struct = yaml.load("""
        testing:
            user:
                key: exists
        testing2:
            inherit: testing
            user:
                key: also_exists""")
        acl = build_acl(struct)
        assert 'key_exists' in acl['testing2']['user']

    def test_basic_inherit_list_type(self):
        struct = yaml.load("""
        testing:
            user:
                key: exists
        testing3:
            user:
                key: super_exists
        testing2:
            inherit:
                - testing
                - testing3
            user:
                key: also_exists""")
        acl = build_acl(struct)
        assert 'key_super_exists' in acl['testing2']['user']



import datetime
import sqlalchemy
import unittest
import json

from flask import Flask, jsonify
from flask.ext.login import LoginManager, current_user, login_user
from sqlalchemy import (Column, create_engine, DateTime, Date, Float,
                        ForeignKey, Integer, Boolean, Unicode, create_engine)
from sqlalchemy.orm.scoping import scoped_session
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base
from pprint import pprint

from lever import API, LeverException, ModelBasedACL, ImpersonateMixin
from lever.mapper import BaseMapper


class FlaskTestBase(unittest.TestCase):
    """Base class for tests which use a Flask application.

    The Flask test client can be accessed at ``self.app``. The Flask
    application itself is accessible at ``self.flaskapp``.

    """

    def setUp(self):
        """Creates the Flask application and the APIManager."""
        # create the Flask application
        app = Flask(__name__)
        app.config['DEBUG'] = True
        app.config['TESTING'] = True
        app.config['SECRET_KEY'] = "testing..."
        del app.logger.handlers[0]
        # sqlalchemy flask
        self.base = declarative_base()
        self.engine = create_engine('sqlite://')
        self.session = scoped_session(sessionmaker(autocommit=False,
                                                   autoflush=False,
                                                   bind=self.engine))
        self.app = app
        self._ctx = self.app.test_request_context()
        self._ctx.push()
        self.client = self.app.test_client()

        # Add an error handler that returns straight LeverException
        # recommendations
        @self.app.errorhandler(LeverException)
        def handler(exc):
            tb = exc.extra.pop('tb', None)
            self.app.logger.debug("Lever Exception Thrown!\n"
                                  "Extra: {0}\nEnd User: {1}"
                                  .format(exc.extra, exc.end_user),
                                    exc_info=True)
            if tb:
                self.app.logger.debug("Traceback from lever exception:"
                                      "\n{0}".format(tb.replace('\\n', '\n')))
            return jsonify(**exc.end_user), exc.code

        return app

    def get(self, uri, status_code, params=None, has_data=True, success=True,
            headers=None):
        if params:
            for p in params:
                if isinstance(params[p], dict) or isinstance(params[p], list):
                    params[p] = json.dumps(params[p])
        if headers is None:
            headers = {}
        response = self.client.get(uri, query_string=params, headers=headers)
        print(response.status_code)
        print(response.data)
        if has_data:
            assert response.data
        j = json.loads(response.data.decode('utf8'))
        pprint(j)
        assert response.status_code == status_code
        if success and status_code == 200:
            assert j['success']
        else:
            assert not j['success']
        return j

    def post(self, uri, status_code, params=None, has_data=True, headers=None,
             success=True, typ='post'):
        if headers is None:
            headers = {}
        response = getattr(self.client, typ)(
            uri,
            data=json.dumps(params),
            headers=headers,
            content_type='application/json')
        print(response.status_code)
        print(response.data)
        j = json.loads(response.data.decode('utf8'))
        pprint(j)
        assert response.status_code == status_code
        if has_data:
            assert response.data
        if success and status_code == 200:
            assert j['success']
        else:
            assert not j['success']
        return j

    def patch(self, uri, status_code, **kwargs):
        return self.post(uri, status_code, typ='patch', **kwargs)

    def put(self, uri, status_code, **kwargs):
        return self.post(uri, status_code, typ='put', **kwargs)

    def delete(self, uri, status_code, **kwargs):
        return self.post(uri, status_code, typ='delete', **kwargs)

    def basic_api(self):
        """ Generates a basic API endpoint with Lever to be tested """
        class Widget(self.base):
            __tablename__ = 'testing'
            id = Column(Integer, primary_key=True)
            name = Column(Unicode, unique=True)
            description = Column(Unicode)
            created_at = Column(DateTime, default=datetime.datetime.utcnow)

            standard_join = ['name', 'created_at', 'id', 'description']


        class WidgetAPI(API):
            model = Widget
            session = self.session

        self.app.add_url_rule('/widget', view_func=WidgetAPI.as_view('widget'))

        return Widget, WidgetAPI

    def provision_single_asset(self):
        widget, api = self.basic_api()
        self.widget_api = api
        self.widget_model = widget
        self.base.metadata.create_all(self.engine)
        obj = widget(name=u'Testing')
        self.session.add(obj)
        self.session.commit()
        return obj

    def provision_many_asset(self):
        widget, api = self.basic_api()
        self.base.metadata.create_all(self.engine)
        objs = []
        for name in [u'sdflgk', u'owuertoi', u'lcxvmnl', u'dlfkjgdsfg']:
            objs.append(widget(name=name))
        self.session.add_all(objs)
        self.session.commit()
        return objs


class TestUserACL(FlaskTestBase):
    """ Used to test full model based ACL system that integrates parent object
    ACL and global user acl roles """

    def setUp(self):
        super(TestUserACL, self).setUp()
        self.lm = LoginManager(self.app)

    def user_api(self):
        self.base = declarative_base(cls=BaseMapper)
        class User(self.base):
            __tablename__ = "user"
            id = Column(Integer, primary_key=True)
            username = Column(Unicode, unique=True)
            description = Column(Unicode)
            password = Column(Unicode)
            admin = Column(Boolean, default=False)
            created_at = Column(DateTime, default=datetime.datetime.utcnow)

            standard_join = ['username', 'created_at', 'id', 'description']
            acl = {'user': set(['view_standard_join']),
                   'anonymous': set(['view_standard_join', 'action_login', 'class_create']),
                   'owner': set(['view_standard_join', 'edit_description']),
                   'admin': set(['delete'])}

            def roles(self, user=current_user):
                if self.id == user.id:
                    return ['owner']
                return []

            def global_roles(self, user=current_user):
                if self.admin:
                    return ['admin', 'user']
                return ['user']

            def login(self, password, user=None):
                if password == self.password:
                    login_user(self)
                    return True
                return False

            def is_authenticated(self):
                return True

            def is_active(self):
                return True

            def is_anonymous(self):
                return False

            def get_id(self):
                return str(self.id)

            @classmethod
            def create(cls, username, password, user=current_user):
                inst = cls(username=username,
                           password=password,
                           description='a new user!')
                self.db.session.add(inst)
                return inst

        # Setup the anonymous user to register a single role
        class AnonymousUser(object):
            id = -100
            gh_token = None
            tw_token = None
            go_token = None

            def is_anonymous(self):
                return True

            def global_roles(self):
                return ['anonymous']

            def is_authenticated(self):
                return False

            def get(self):
                return self
        self.lm.anonymous_user = AnonymousUser

        # setup login manager stuff
        @self.lm.user_loader
        def user_loader(id):
            try:
                return self.session.query(User).filter_by(id=id).one()
            except sqlalchemy.orm.exc.NoResultFound:
                return None

        class UserAPI(ModelBasedACL, ImpersonateMixin, API):
            model = User
            session = self.session
            user_model = User

        self.app.add_url_rule('/user', view_func=UserAPI.as_view('user'))
        self.user_model = User
        self.user_api = UserAPI

    def login(self, username):
        login_user(self.User.query.filter_by(username=username).one())

    def provision_users(self):
        """Creates the database, the Flask application, and the APIManager."""
        people = []
        for u in [u'mary', u'lucy', u'katy', u'john']:
            user = self.user_model(username=u, password=u'testing')
            people.append(user)

        # make an admin user
        self.admin = self.user_model(username=u'admin', password=u'testing', admin=True)
        people.append(self.admin)

        self.session.add_all(people)
        self.session.commit()
        return people



import unittest
import types
import datetime

from flask import Flask
from pprint import pprint
from sqlalchemy import (Column, create_engine, DateTime, Date, Float,
                        ForeignKey, Integer, Boolean, Unicode, create_engine)

from lever import API, preprocess, postprocess, ModelBasedACL, ImpersonateMixin
from lever.tests.model_helpers import FlaskTestBase, TestUserACL


class ProcessTests(unittest.TestCase):
    """ Ensures our metaclasses and decorators operate as we want for assigning
    preprocessors and postprocessors """
    def test_basic_preprocess(self):
        class APIAwesome(API):
            @preprocess(method='post')
            def preprocess_those(self):
                pass

            @preprocess(action='something')
            def preprocess_that(self):
                pass

        assert isinstance(APIAwesome._pre_method['post'][0],
                          types.FunctionType)
        assert isinstance(APIAwesome._pre_action['something'][0],
                          types.FunctionType)

    def test_inheritence_mixins(self):
        class APIParent(object):
            @preprocess(method='post')
            def preprocess_those(self):
                pass
        class APIAwesome(API, APIParent):
            pass

        assert isinstance(APIAwesome._pre_method['post'][0],
                          types.FunctionType)

    def test_inheritence(self):
        class APIParent(API):
            @preprocess(method='post')
            def preprocess_those(self):
                pass
        class APIAwesome(APIParent):
            pass

        assert isinstance(APIAwesome._pre_method['post'][0],
                          types.FunctionType)

    def test_inheritence_reversal(self):
        class APIParent(API):
            pass
        class APIAwesome(APIParent):
            @preprocess(method='post')
            def preprocess_those(self):
                pass

        assert isinstance(APIAwesome._pre_method['post'][0],
                          types.FunctionType)

    def test_multi_preprocess(self):
        class APIAwesome(API):
            @preprocess(method=['post', 'get'])
            def preprocess_those(self):
                pass

            @preprocess(action=['create', 'other'])
            def preprocess_that(self):
                pass

        assert isinstance(APIAwesome._pre_method['post'][0],
                          types.FunctionType)
        assert isinstance(APIAwesome._pre_method['get'][0],
                          types.FunctionType)
        assert isinstance(APIAwesome._pre_action['other'][0],
                          types.FunctionType)
        assert isinstance(APIAwesome._pre_action['create'][0],
                          types.FunctionType)

    def test_basic_postprocess(self):
        class APIAwesome(API):
            @postprocess(method='post')
            def preprocess_those(self):
                pass

            @postprocess(action='something')
            def preprocess_that(self):
                pass

        assert isinstance(APIAwesome._post_method['post'][0],
                          types.FunctionType)
        assert isinstance(APIAwesome._post_action['something'][0],
                          types.FunctionType)

    def test_multi_postprocess(self):
        class APIAwesome(API):
            @postprocess(method=['post', 'get'])
            def preprocess_those(self):
                pass

            @postprocess(action=['create', 'other'])
            def preprocess_that(self):
                pass

        assert isinstance(APIAwesome._post_method['post'][0],
                          types.FunctionType)
        assert isinstance(APIAwesome._post_method['get'][0],
                          types.FunctionType)
        assert isinstance(APIAwesome._post_action['other'][0],
                          types.FunctionType)
        assert isinstance(APIAwesome._post_action['create'][0],
                          types.FunctionType)

    def test_none(self):
        class APIAwesome(API):
            pass

        assert APIAwesome._pre_method == {}
        assert APIAwesome._pre_action == {}


class TestProcessorUsage(FlaskTestBase):
    """ These tests ensure that preprocessors and postprocessors are getting
    called when they should be """

    def test_methods_preprocess(self):
        for meth in ['post', 'get', 'delete', 'put']:
            class APIAwesome(API):
                @preprocess(method=meth)
                def preprocessor_one(self):
                    raise SyntaxError  # pick an obscure one to catch..

            inst = APIAwesome()
            self.assertRaises(SyntaxError, getattr(inst, meth))

    def test_methods_postprocess(self):
        obj = self.provision_single_asset()
        data = [('post', {'name': 'test'}),
                ('get', {}),
                ('put', {'id': obj.id, 'name': 'test2'}),
                ('delete', {'id': obj.id})]
        for meth, vals in data:
            class APIAwesome(self.widget_api):
                @postprocess(method=meth)
                def postprocess_one(self, retval):
                    raise SyntaxError  # pick an obscure one to catch..

            self.app.add_url_rule('/' + meth, view_func=APIAwesome.as_view(meth))

        for meth, vals in data:
            self.assertRaises(SyntaxError, getattr(self, meth), meth, 500, params=vals)


class TestAPICreation(FlaskTestBase):
    def test_create_bad_pkey(self):
        """ ensure that exception is thrown for invalid primary_key """
        class Testing(self.base):
            __tablename__ = "testing_table"
            bad_id = Column(Integer, primary_key=True)

        class UserAPI(API):
            model = Testing
            session = self.session

        t = UserAPI()
        self.assertRaises(AttributeError, lambda: t.pkey)


class TestGet(FlaskTestBase):
    """ Test facets of our get method """

    def test_get_pkey(self):
        obj = self.provision_single_asset()
        d = self.get('widget', 200, {'id': obj.id})
        assert len(d['objects']) > 0
        assert d['objects'][0]['id'] == obj.id

    def test_many_query(self):
        self.provision_many_asset()
        d = self.get('widget', 200)
        assert len(d['objects']) >= 4


class TestPut(FlaskTestBase):
    """ Test facets of our get method """
    def test_update(self):
        """ can we change an object """
        obj = self.provision_single_asset()
        test_string = "testing this thing"
        p = {'id': obj.id, 'description': test_string}
        self.put('widget', 200, params=p)
        self.session.refresh(obj)
        assert obj.description == test_string

    def test_cant_find(self):
        self.basic_api()
        self.base.metadata.create_all(self.engine)
        ret = self.put('widget', 404, params={'id': 123})
        assert 'not be found' in ret['message']

    def test_cant_find_invalid_key(self):
        self.basic_api()
        self.base.metadata.create_all(self.engine)
        ret = self.put('widget', 404, params={'tid': 123})
        assert 'any object to update' in ret['message']


class TestDelete(FlaskTestBase):
    def test_delete(self):
        """ can we delete an object """
        obj = self.provision_single_asset()
        obj_id = obj.id
        self.delete('widget', 200, params={'id': obj_id})
        obj = self.session.query(self.widget_model).filter_by(id=obj_id).first()
        assert obj is None

    def test_cant_find_put_delete(self):
        self.basic_api()
        self.base.metadata.create_all(self.engine)
        ret = self.delete('widget', 404, params={'id': 123})
        assert 'Object could not be found' in ret['message']

    def test_cant_find(self):
        self.basic_api()
        self.base.metadata.create_all(self.engine)
        ret = self.delete('widget', 404, params={'tid': 123})
        assert 'object to delete' in ret['message']


class TestPost(FlaskTestBase):
    def test_create_dup(self):
        """ make a duplicate entry and fail """
        obj = self.provision_single_asset()
        p = self.post('widget', 409, params={'name': u'Testing'})
        assert 'duplicate value already' in p['message']

    def test_create_new(self):
        """ try creating a new object """
        self.basic_api()
        self.base.metadata.create_all(self.engine)
        p = self.post('widget', 200, params={'name': u'Testing'})
        assert p['objects'][0]['name'] == 'Testing'

    def test_bad_action(self):
        obj = self.provision_single_asset()
        p = {'id': obj.id, '__action': 'dsflgjksdfglk'}
        ret = self.post('widget', 400, params=p)
        assert 'missing key' in ret['message']


class TestSearch(FlaskTestBase):
    """ Run a bunch of positive and negative tests on our searching system """
    def test_filter_by(self):
        obj = self.provision_single_asset()
        d = self.get('widget', 200, {'__filter_by': {'name': obj.name}})
        assert len(d['objects']) > 0
        assert d['objects'][0]['name'] == obj.name

    def test_single_failure(self):
        self.basic_api()
        self.base.metadata.create_all(self.engine)
        ret = self.get('widget', 404, params={'__one': True})
        assert 'not be found' in ret['message']

    def test_query_bad_param_filter_by(self):
        self.basic_api()
        ret = self.get('widget', 400, params={'__filter_by': {'sdflgj': True}})
        assert 'invalid field' in ret['message']

    def test_query_filter(self):
        obj = self.provision_single_asset()
        ret = self.get('widget', 200,
                       params={'__filter': [
                           {'val': 'Testing', 'name': 'name', 'op': 'eq'}]})
        assert ret['objects'][0]['name'] == 'Testing'

    def test_query_filter_field(self):
        """ test filtering by comparing different fields to eachother """
        obj = self.provision_single_asset()
        # TODO: Write a positive test for this
        ret = self.get('widget', 200,
                       params={'__filter': [
                           {'field': 'created_at', 'name': 'name', 'op': 'eq'}]})
        assert len(ret['objects']) == 0

    def test_query_bad_param_filter(self):
        self.basic_api()
        ret = self.get('widget', 400,
                       params={'__filter': [
                           {'val': True, 'name': 'dsflgjsdflgk', 'op': 'eq'}]})
        assert 'invalid field' in ret['message']

    def test_query_bad_param_op(self):
        self.basic_api()
        ret = self.get('widget', 400,
                 params={'__filter': [
                     {'val': True, 'name': 'name', 'op': 'fake'}]})
        assert 'operator specified in' in ret['message']

    def test_query_missing_param(self):
        self.basic_api()
        ret = self.get('widget', 400,
                 params={'__filter': [
                     {'val': True, 'name': 'name', 'op2': 'fake'}]})
        assert 'missing required arguments' in ret['message']

    def test_query_bad_param_count(self):
        self.basic_api()
        ret = self.get('widget', 400,
                 params={'__filter': [{'name': 'name', 'op': '=='}]})
        assert 'argument count' in ret['message']

    def test_order_by(self):
        self.provision_many_asset()
        ret = self.get('widget', 200, params={'__order_by': ['id']})
        assert len(ret['objects']) >= 4
        assert ret['objects'][0]['id'] < ret['objects'][1]['id']
        assert ret['objects'][2]['id'] < ret['objects'][3]['id']

    def test_order_by_desc(self):
        self.provision_many_asset()
        ret = self.get('widget', 200, params={'__order_by': ['-id']})
        assert len(ret['objects']) >= 4
        assert ret['objects'][0]['id'] > ret['objects'][1]['id']
        assert ret['objects'][2]['id'] > ret['objects'][3]['id']

    def test_order_by_bad_key(self):
        self.basic_api()
        ret = self.get('widget', 400, params={'__order_by': ['dflgjksdfgl']})
        assert 'Order_by operator' in ret['message']


class TestLogin(TestUserACL):
    """ Tests abilities of the User ACL mixin class """
    def test_login(self):
        """ can we login with a patch action """
        self.user_api()
        self.base.metadata.create_all(self.engine)
        people = self.provision_users()
        p = {'__action': 'login', 'id': people[0].id, 'password': "testing"}
        self.post('user', 200, params=p)

    def test_delete_fail(self):
        """ will delete fail with bad permissions? """
        self.user_api()
        self.base.metadata.create_all(self.engine)
        people = self.provision_users()
        p = {'id': people[2].id}
        self.delete('user', 403, params=p)

    def test_impersonate(self):
        """ can the admin properly impersonate someone for a create?"""
        self.user_api()
        class Widget(self.base):
            __tablename__ = 'testing'
            id = Column(Integer, primary_key=True)
            name = Column(Unicode, unique=True)
            owner = Column(Unicode)
            created_at = Column(DateTime, default=datetime.datetime.utcnow)

            acl = {'admin': set(['class_create_other', 'action_create'])}
            standard_join = ['name', 'created_at', 'id']

            @classmethod
            def create(cls, name, user=None):
                inst = cls(name=name)
                if user:
                    self.owner = user.username
                self.session.add(inst)
                return inst


        class WidgetAPI(ImpersonateMixin, ModelBasedACL, API):
            model = Widget
            session = self.session
            create_method = "create"
            user_model = self.user_model


        self.app.add_url_rule('/widget', view_func=WidgetAPI.as_view('widget'))
        self.base.metadata.create_all(self.engine)
        people = self.provision_users()
        p = {'__action': 'login', 'id': self.admin.id, 'password': "testing"}
        self.post('user', 200, params=p)
        p = {'__user_id': people[2].id, 'name': 'testing'}
        self.post('widget', 200, params=p)



""" This module allows you to build acl roles for various resources using
convenient tools like inheritence between roles and inheritence between
resources, as well as common prefix lists to prevent re-trying prefixes all
the time. For instance, I might have an object called Book that I want to make
ACL records for. Well we might do it like such:

book:
    user:
        view:
            - pages
            - contents
        add:
            - comments
    author:
        inherit: user
        edit:
            - pages
            - contents

This would all be compiled into a structure like this:
{'book':
    'user': ['view_pages', 'view_contents', 'add_comments'],
    'author': ['view_pages', 'view_contents', 'add_comments',
               'edit_pages', 'edit_contents']
}
This object is ready to be used in defining object ACL rules with lever """
import six


def inherit_dict(*args):
    """ Joines together multiple dictionaries left to right """
    ret = {}
    for arg in args:
        ret.update(arg)
    return ret


def build_acl(structure):
    acl = {}
    # do a run to compile all dictionaries into lists and merge the lists to
    # createa a role list of keys
    for typ, roles in six.iteritems(structure):
        acl.setdefault(typ, {})  # set the key to a dict
        for role, keys in six.iteritems(roles):
            if role in ['inherit', 'virtual']:  # skip inherit commands
                continue
            acl[typ].setdefault(role, set())
            if isinstance(keys, list):
                acl[typ][role] |= set(keys)
            elif isinstance(keys, dict):
                for key, val in six.iteritems(keys):
                    # skip inheritence clauses, handled later
                    if key == "inherit":
                        continue
                    if isinstance(val, list):  # if its a list, prepend
                        acl[typ][role] |= set([key + "_" + v for v in val])
                    elif isinstance(val, six.string_types):
                        acl[typ][role].add(key + "_" + val)
                    else:
                        raise Exception(
                            "Type {0} not supported".format(type(val)))

    def compile(acl, typ, stack, compiled):
        """ Compiles a specific type, but calls itself recursively to compile
        it's own dependencies. Detects inheritence loops by checking the stack.
        """
        if typ in compiled:
            return

        for role, keys in six.iteritems(structure[typ]):
            if role == 'inherit':
                if not isinstance(keys, list):  # allow single entry, or list
                    keys = [keys]
                for inh_type in keys:
                    # run inheritence from another type
                    if inh_type not in structure:
                        raise KeyError(
                            "Unable to inherit from type {0} for type {1}, "
                            "doesn't exist".format(inh_type, typ))
                    if inh_type in stack:
                        raise Exception(
                            "Looping inheritence detected! Type {0} tried to "
                            "inherit from type {1} which was called to compile"
                            "{0}!".format(typ, inh_type))
                    if inh_type not in compiled:
                        compile(acl, inh_type, stack + [inh_type], compiled)
                    for inh_role in acl[inh_type]:
                        acl[typ].setdefault(inh_role, set())
                        acl[typ][inh_role] |= acl[inh_type][inh_role]

        for role, keys in six.iteritems(structure[typ]):
            if isinstance(keys, dict):
                for key, val in six.iteritems(keys):
                    if key == "inherit":
                        # allow single entry or list
                        if not isinstance(val, list):
                            val = [val]
                        for inh in val:
                            try:
                                acl[typ][role] |= acl[typ][inh]
                            except KeyError:
                                raise KeyError(
                                    "Unable to inherit from role {0} for role "
                                    "{1} on type {2}".format(inh, role, typ))
        compiled.append(typ)

    compiled = []
    for typ in structure:
        compile(acl, typ, [typ], compiled)

    # inheritence pass. Allows a type to inherit from another type
    for typ, roles in six.iteritems(structure):
        for role, keys in six.iteritems(roles):
            if role == 'virtual':  # remove virtual types
                del acl[typ]

    return acl



from flask.views import MethodView, MethodViewType
from flask import jsonify, current_app, request

import json
import six
import sqlalchemy
import sys
import calendar
import datetime
import traceback


class LeverException(Exception):
    """ Lever handles many common errors that occur in the API and raises this
    exception with recommended return information.

    An attribute 'code' carries an http return code.

    End_user carries a dictionary of information that could be helpful to
    return to the user. End user dictionary will always include a key message,
which will mirror the message text of the exception, but can contain more
information when applicable information is available. The original stack trace
will always be passed back on the Exception.

    In addition to the end_user dictionary, there is an extra dictionary which
    will contain potentially helpful debugging information, but may be unsafe
    to return to the User.

    These exceptions can be raised by your custom PATCH methods and will be
    handled as expected. Anything raised by custom methods that isn't a
    KeyError, AttributeError, or child of LeverException will cause a 500.  """
    code = 500

    def __init__(self, message, code=None, end_user=None, extra=None):
        Exception.__init__(self, message)
        if code is not None:
            self.code = code
        self.message = message
        if end_user is None:
            self.end_user = {}
        else:
            self.end_user = end_user
        if extra is None:
            self.extra = {}
        else:
            self.extra = extra
        self.end_user.update({'message': message})


class LeverSyntaxError(LeverException):
    code = 400


class LeverNotFound(LeverException):
    code = 404
    pass


class LeverServerError(LeverException):
    code = 500


OPERATORS = {
    # Operators which accept a single argument.
    'is_null': lambda f: f == None,
    'is_not_null': lambda f: f != None,
    # TODO what are these?
    'desc': lambda f: f.desc,
    'asc': lambda f: f.asc,
    # Operators which accept two arguments.
    '==': lambda f, a: f == a,
    'eq': lambda f, a: f == a,
    'equals': lambda f, a: f == a,
    'equal_to': lambda f, a: f == a,
    '!=': lambda f, a: f != a,
    'ne': lambda f, a: f != a,
    'neq': lambda f, a: f != a,
    'not_equal_to': lambda f, a: f != a,
    'does_not_equal': lambda f, a: f != a,
    '>': lambda f, a: f > a,
    'gt': lambda f, a: f > a,
    '<': lambda f, a: f < a,
    'lt': lambda f, a: f < a,
    '>=': lambda f, a: f >= a,
    'ge': lambda f, a: f >= a,
    'gte': lambda f, a: f >= a,
    'geq': lambda f, a: f >= a,
    '<=': lambda f, a: f <= a,
    'le': lambda f, a: f <= a,
    'lte': lambda f, a: f <= a,
    'leq': lambda f, a: f <= a,
    'ilike': lambda f, a: f.ilike(a),
    'like': lambda f, a: f.like(a),
    'in': lambda f, a: f.in_(a),
    'not_in': lambda f, a: ~f.in_(a),
}


class preprocess(object):
    """ Simple decorator that sets special attributes on decorated methods.
    These attributes get picked up by our __new__ method and assign
    preprocessors for specific actions or methods """
    def __init__(self, method=None, action=None):
        self.action = action
        self.method = method

    def __call__(self, f):
        f._pre_action = self.action
        f._pre_method = self.method
        return f

class postprocess(preprocess):
    def __call__(self, f):
        f._post_action = self.action
        f._post_method = self.method
        return f


class ImpersonateMixin(object):
    """ Allows ACL lookups from a user object. Expects current_user attribute
    to be set to an application LocalProxy that yeilds the user object for the
    request. This is the most common ACL system used with Lever. """
    from flask.ext.login import current_user
    user_model = None

    @preprocess('post')
    def impersonate(self):
        if self.user_model is None:
            raise Exception("user_model must be defined to lookup users")
        # check to ensure the user can create for others if requested
        username = self.params.pop('__username', None)
        userid = self.params.pop('__user_id', None)
        if userid or username:
            query = self.session.query(self.user_model)
            if userid:
                query = query.filter_by(id=userid)
            if username:
                query = query.filter_by(username=username)
            user = query.one()
            assert self.can_cls('class_create_other'), "Cant create for other users"
            self.params['user'] = user
        else:
            assert self.can_cls('class_create'), "Cant create that object"
            self.params['user'] = self.current_user.get()

    def can_cls(self, action):
        # don't pass in impersonated user objects...
        dct = dict((k, v) for k, v in self.params.items() if k != 'user')
        return self.model.can_cls(action, **dct)


class ModelBasedACL(object):
    """ Executes ACL methods on the model directly to determine ability to
    execute requested action """
    def can(self, obj, action):
        return obj.can(action)

    def can_cls(self, action):
        return self.model.can_cls(action, **self.params)


class APIMeta(MethodViewType):
    def __init__(mcs, name, bases, dct):
        MethodViewType.__init__(mcs, name, bases, dct)
        types = ['_pre_method', '_post_method', '_pre_action', '_post_action']
        for key in types:
            setattr(mcs, key, {})
        attrs = list(six.itervalues(dct))
        for base in bases:
            attrs.extend(list(six.itervalues(base.__dict__)))
        for attr in attrs:
            for key in types:
                val = getattr(attr, key, None)
                if val:
                    if not isinstance(val, (list, tuple)):
                        val = (val, )
                    for method in val:
                        getattr(mcs, key).setdefault(method, []).append(attr)


class API(six.with_metaclass(APIMeta, MethodView)):
    """ The main method that underlies Lever. Create a new class that inherits
    from this and set the session to your database session, model to the model
    that you're wrapping and basic functionality should be good to go. """
    max_pg_size = 100
    # Defines the maximum pagination size that the client can select
    pkey_val = 'id'
    # defines the primary key for your model. this value will be expected on
    # get and updates
    create_method = '__init__'
    params = {}
    session = None
    # The database session from SQLAlchemy

    @property
    def pkey(self):
        try:
            return getattr(self.model, self.pkey_val)
        except AttributeError:
            raise AttributeError(
                'Invalid primary key defined for model {0}'
                .format(self.model.__class__.__name__))

    def dispatch_request(self, *args, **kwargs):
        meth = getattr(self, request.method.lower(), None)
        # if the request method is HEAD and we don't have a handler for it
        # retry with GET
        if meth is None and request.method == 'HEAD':
            meth = getattr(self, 'get', None)
        assert meth is not None, 'Unimplemented method %r' % request.method

        try:
            return meth(*args, **kwargs)
        except Exception:
            # capture the info
            info = sys.exc_info()

        extra = {'original_exc': str(info[1]),
                 'original_exc_type': str(info[0]),
                 'tb': "".join(traceback.format_tb(info[2])),
                 'params': self.params}
        end_user = {'success': False}
        code = None

        try:
            # reraise to handle differnet exceptions differently
            six.reraise(*info)
        # Common API errors
        except KeyError as e:
            msg = 'Incorrect syntax or missing key ' + str(e)
            code = 400
            extra = {'key': str(e)}
        except AttributeError:
            msg = 'Incorrect syntax or missing key'
            code = 400
        except AssertionError:
            msg = "You don't have permission to do that"
            code = 403
        except LeverException as e:
            # add in the default params, but let the manually raised exception
            # take precedence
            end_user.update(e.end_user)
            extra.update(e.extra)
            e.extra = extra
            e.end_user = end_user
            six.reraise(LeverException, e, tb=sys.exc_info()[2])

        # SQLAlchemy exceptions
        except sqlalchemy.orm.exc.NoResultFound:
            msg = 'Object could not be found'
            code = 404
        except sqlalchemy.orm.exc.MultipleResultsFound:
            msg = 'Only one result requested, but MultipleResultsFound'
            code = 400
        except sqlalchemy.exc.IntegrityError:
            msg = "A duplicate value already exists in the database"
            code = 409
        except sqlalchemy.exc.InvalidRequestError:
            msg = "Client programming error, likely invalid search sytax used."
            code = 400
        except sqlalchemy.exc.DataError:
            msg = "ORM returned invalid data for an argument"
            code = 400
        except sqlalchemy.exc.SQLAlchemyError:
            msg = "An unknown database operations error has occurred"
            code = 402
        except:
            raise

        # get our exception info and try to extract extra information out of it
        exc = LeverException(msg, code=code, extra=extra, end_user=end_user)
        six.reraise(LeverException, exc, tb=info[2])

    def get_obj(self):
        pkey = self.params.pop(self.pkey_val, None)
        if pkey:  # if a int primary key is passed
            return self.session.query(self.model).filter(self.pkey == pkey).one()
        return False

    def can(self, obj, action):
        """ This function should parse the current parameters to gain parent
        information for properly running can_cls on the model this API wraps
        """
        return True

    def can_cls(self, action):
        """ This function should parse the current parameters to gain parent
        information for properly running can_cls on the model this API wraps
        """
        return True

    def get(self):
        """ Retrieve an object from the database """
        # convert args to a real dictionary that can be popped
        self.params = dict((one, two) for one, two in six.iteritems(request.args))
        for method in self._pre_method.get('get', []):
            method(self)
        join = self.params.pop('join_prof', 'standard_join')
        obj = self.get_obj()
        if obj:  # if a int primary key is passed
            assert self.can(obj, 'view_' + join), "Can't view that object with join " + join
            retval = dict(success=True, objects=[get_joined(obj, join)])
        else:
            query = self.search()
            one = self.params.pop('__one', None)
            if one:
                query = [query.one()]
            else:
                query = self.paginate(query=query)
            for obj in query:
                assert self.can(obj, 'view_' + join), "Can't view that object with join " + join
            retval = dict(success=True, objects=get_joined(query, join))

        for method in self._post_method.get('get', []):
            method(self, retval)
        return jsonify(**retval)

    def post(self):
        """ Create a new object """
        self.params = request.get_json(silent=True) or {}
        for method in self._pre_method.get('post', []):
            method(self)
        action = self.params.pop('__action', None)
        cls = self.params.pop('__cls', None)
        if not action:
            cls = True
            action = self.create_method

        if not cls:
            obj = self.get_obj()
            if not obj:
                raise LeverNotFound(
                    "Could not find any object to perform an action on")
            assert self.can(obj, 'action_' + action), "Cant perform action " + action
        else:
            assert self.can_cls('action_' + action), "Can't perform cls action " + action
            obj = self.model

        retval = {}
        try:
            if action == '__init__':
                obj = obj(**self.params)
                self.session.add(obj)
                self.session.flush()
                ret = {'objects': [get_joined(obj)]}
            else:
                ret = getattr(obj, action)(**self.params)
                if isinstance(ret, self.model):
                    ret = {'objects': [get_joined(ret)]}
        except TypeError as e:
            if 'argument' in str(e):
                msg = ("Wrong number of arguments supplied for action {0}."
                       .format(action))
                six.reraise(LeverSyntaxError, msg, tb=sys.exc_info()[2])
            else:
                raise

        if ret is None or ret is True:
            retval['success'] = True
        elif ret is False:
            retval['success'] = False
        else:
            retval['success'] = True
            retval.update(ret)

        self.session.commit()

        for method in self._post_method.get('post', []):
            method(self, retval)
        try:
            return jsonify(**retval)
        except (LeverException, KeyError, AttributeError):
            raise
        except Exception as e:
            if 'not JSON serializable' in str(e):
                raise LeverServerError(
                    "Response was not JSON serializable, patch method returned"
                    " invalid data.",
                    extra={'retval': str(retval)},
                    end_user={'method': action})

    def create_hook(self):
        """ Does logic required for checking permissions on a create action """
        pass

    def put(self):
        """ Updates an objects values """
        self.params = request.get_json(silent=True)
        for method in self._pre_method.get('put', []):
            method(self)
        if not self.params:
            raise LeverSyntaxError("To update, values must be specified")
        obj = self.get_obj()
        if not obj:
            raise LeverNotFound("Could not find any object to update")

        # updates all fields if data is provided, checks acl
        for key, val in six.iteritems(self.params):
            current_app.logger.debug(
                "Updating value for '{0}' to '{1}'".format(key, val))
            assert self.can(obj, 'edit_' + key), "Can't edit key {0} on type {1}"\
                .format(key, self.model.__name__)
            setattr(obj, key, val)

        self.session.commit()
        retval = {'success': True}
        for method in self._post_method.get('put', []):
            method(self, retval)

        return jsonify(**retval)

    def delete(self):
        self.params = request.get_json(silent=True)
        for method in self._pre_method.get('delete', []):
            method(self)
        if not self.params:
            raise LeverSyntaxError("To delete, values must be specified")

        obj = self.get_obj()
        if not obj:
            raise LeverNotFound("Could not find any object to delete")
        assert self.can(obj, 'delete'), "Can't delete that object"
        self.session.delete(obj)
        self.session.commit()

        retval = {'success': True}
        for method in self._post_method.get('delete', []):
            method(self, retval)
        return jsonify(**retval)

    def paginate(self, query=None):
        """ Sets limit and offset values on a query object based on arguments,
        and limited by class settings """
        if not query:
            query = self.session.query(self.model)
        pg_size = self.params.get('pg_size')
        # don't do any pagination if we don't have a max page size and no
        # pagination is requested
        if pg_size is None and self.max_pg_size is None:
            return query
        elif pg_size is None:  # default to max_pg_size
            pg_size = self.max_pg_size
        pg_size = min(pg_size, self.max_pg_size)  # limit their option to max
        page = self.params.get('pg', 1)
        return query.offset((page - 1) * pg_size).limit(pg_size)

    def search(self, query=None):
        """ Handles arguments __filter_by, __filter, and __order_by by
        modifying the query parameters before execution """
        if query is None:
            query = self.session.query(self.model)

        filters = self.params.pop('__filter', None)
        try:
            if filters:
                # it's a json encoded parameter to get
                if isinstance(filters, six.string_types):
                    filters = safe_json(filters)
                for op in filters:
                    args = []
                    args.append(getattr(self.model, op['name']))
                    if 'val' in op:
                        args.append(op['val'])
                    if 'field' in op:
                        args.append(getattr(self.model, op['field']))
                    operator = OPERATORS.get(op['op'])
                    if operator is None:
                        raise LeverSyntaxError("Invalid operator specified in filter arguments")
                    func = operator(*args)
                    query = query.filter(func)
        except AttributeError:
            current_app.logger.debug("Attribute filter error", exc_info=True)
            raise LeverSyntaxError(
                'Filter operator "{0}" accessed invalid field'
                .format(op))
        except KeyError:
            current_app.logger.debug("Key filter error", exc_info=True)
            raise LeverSyntaxError(
                'Filter operator "{0}" was missing required arguments'
                .format(op))
        except TypeError:
            current_app.logger.debug("Argument count error", exc_info=True)
            raise LeverSyntaxError(
                'Incorrect argument count for requested filter operation'
                .format(op))

        order_by = self.params.pop('__order_by', None)
        try:
            if order_by:
                # it's a json encoded parameter to get
                if isinstance(order_by, six.string_types):
                    order_by = safe_json(order_by)
                for key in order_by:
                    if key.startswith('-'):
                        base = getattr(self.model, key[1:]).desc()
                    else:
                        base = getattr(self.model, key)
                    query = query.order_by(base)
        except AttributeError:
            raise LeverSyntaxError(
                'Order_by operator "{0}" accessed invalid field'
                .format(key))

        filter_by = self.params.pop('__filter_by', None)
        if filter_by:
            # it's a json encoded parameter to get
            if isinstance(filter_by, six.string_types):
                filter_by = safe_json(filter_by)
            for key, value in six.iteritems(filter_by):
                try:
                    query = query.filter_by(**{key: value})
                except sqlalchemy.exc.InvalidRequestError:
                    raise LeverSyntaxError(
                        'Filter_by key "{0}" accessed invalid field'
                        .format(key))

        return query

    @classmethod
    def register(cls, mod, url):
        """ Registers the API to a blueprint or application """
        symfunc = cls.as_view(cls.__name__)
        mod.add_url_rule(url,
                         view_func=symfunc,
                         methods=['GET', 'POST', 'PUT', 'DELETE'])


def get_joined(obj, join_prof="standard_join"):
    # If it's a list, join each of the items in the list and return
    # modified list
    if isinstance(obj, (sqlalchemy.orm.Query,
                        sqlalchemy.orm.collections.InstrumentedList,
                        list)):
        lst = []
        for item in obj:
            lst.append(get_joined(item, join_prof=join_prof))
        return lst

    # split the join list into it's compoenents, obj to be removed, sub
    # object join data, and current object join values
    if isinstance(join_prof, six.string_types):
        join = getattr(obj, join_prof)
    else:
        join = join_prof

    remove = []
    sub_obj = []
    join_keys = []
    for key in join:
        if isinstance(key, six.string_types):
            if key.startswith('-'):
                remove.append(key[1:])
            else:
                join_keys.append(key)
        else:
            sub_obj.append(key)

    include_base = False
    try:
        join_keys.remove('__dont_mongo')
    except ValueError:
        include_base = True
    # run the primary object join
    join_vals = jsonize(obj, join_keys, raw=True)
    # catch our special config key
    if include_base:
        # first we get the names of all the columns on your model
        columns = [c.key for c in sqlalchemy.orm.class_mapper(obj.__class__).columns]
        # then we return their values in a dict
        dct = dict((c, getattr(obj, c)) for c in columns)
        # Remove keys from the bson that the join prefixes with a -
        for key in remove:
            dct.pop(key, None)
        dct.update(join_vals)
    else:
        dct = join_vals
    dct['_cls'] = obj.__class__.__name__

    # run all the subobject joins
    for conf in sub_obj:
        key = conf.get('obj')
        # allow the conf dictionary to specify a join profiel
        prof = conf.get('join_prof', "standard_join")
        subobj = getattr(obj, key)
        if subobj is not None:
            dct[key] = get_joined(subobj, join_prof=prof)
        else:
            current_app.logger.info(
                "Attempting to access attribute {} from {} resulted in {} "
                "type".format(key, type(obj), subobj))
            dct[key] = subobj
    return dct


def safe_json(json_string):
    try:
        return json.loads(json_string)
    except Exception as e:
        raise LeverSyntaxError(
            "Error decoding JSON parameters. Original exception was {}"
            .format(e))


def jsonize(obj, args, raw=False):
    """ Used to join attributes or functions to an objects json
    representation.  For passing back object state via the api """
    dct = {}
    for key in args:
        attr = getattr(obj, key)
        # if it's a callable function call it, then do the parsing below
        if callable(attr):
            try:
                attr = attr()
            except TypeError:
                current_app.logger.warn(
                    "{0} callable requires argument on obj {1}"
                    .format(str(attr), obj.__class__.__name__))
                continue
        # convert datetime to seconds since epoch, much more universal
        if isinstance(attr, datetime.datetime):
            attr = calendar.timegm(attr.utctimetuple()) * 1000
        # don't convert these common types to strings
        elif (isinstance(attr, bool) or
              isinstance(attr, int) or
              isinstance(attr, dict) or
              attr is None or
              isinstance(attr, list)):
            pass
        # convert set to a dictionary for easy conditionals
        elif isinstance(attr, set):
            attr = dict((x, True) for x in attr)
        else:  # if we don't know what it is, stringify it
            attr = str(attr)

        dct[key] = attr

    if raw:
        return dct
    else:
        return json.dumps(dct)



#!/usr/bin/env python

from setuptools import setup, find_packages

requires = ['flask',
            'six',
            'sqlalchemy']

setup(name='lever',
      version='0.2.0',
      description='A tool for exposing SQLAlchemy models in Flask via REST',
      author='Isaac Cook',
      author_email='isaac@crowdlink.io',
      install_requires=requires,
      url='http://www.python.org/sigs/distutils-sig/',
      packages=find_packages()
      )



__version__ = '0.2'

# make the following names available as part of the public API
from .base import (API, LeverException, get_joined, LeverSyntaxError, jsonize,
                   preprocess, postprocess, ModelBasedACL, ImpersonateMixin)
from .acl import build_acl



from flask import current_app
from flask.ext.login import current_user
from flask.ext.sqlalchemy import BaseQuery

import six


class BaseMapper(object):
    """ The base model instance for all model. Provides lots of useful
    utilities in addition to access control logic and serialization helpers """

    # Allows us to run query on the class directly, instead of through a
    # session
    query_class = BaseQuery
    query = None

    # Access Control Methods
    # =========================================================================
    def roles(self, user=current_user):
        """ This should be overriden to use logic for determining roles on an
        instance of the class. This should include inheriting parent roles from
        p_roles function. """
        return []

    @classmethod
    def p_roles(self, **parents):
        """ Determines roles to be gained from parent objects. Usually uses the
        _inherit_roles helper function to prefix all parent roles with their
        class name """
        return []

    def can(self, action, user=current_user):
        """ Can the user perform the action needed on this object instance?
        Checks for the desired key in a list of allowed action keys. """
        keys = self.user_acl(user=user)
        return action in keys

    @classmethod
    def can_cls(cls, action, user=current_user, **parents):
        """ Similar to can, except does not include instance specific roles.
        Intended to be used to determine if pre-creation events can occur, such
        as create or create_other. Requires the data on parents to be passed in
        via keyword arguments to determine parent roles"""
        return action in cls._role_mix(cls.p_roles(**parents) + user.global_roles())

    def user_acl(self, user=current_user):
        """ A list of access keys the user has with context to the current
        object """
        roles = self.roles(user=user) + user.global_roles()
        return self._role_mix(roles)

    @classmethod
    def _role_mix(cls, roles):
        """ A utility that takes a list of roles and returns a set of allowed
        actions that was determined by those roles """
        allowed = set()
        for role in roles:
            allowed |= cls.acl.get(role, set())
        current_app.logger.debug((roles, allowed))
        return allowed

    @classmethod
    def _inherit_roles(cls, user=current_user, **kwargs):
        """ a utility method that prefixes the roles of parents """
        r = []
        for prefix, obj in six.iteritems(kwargs):
            r += [prefix + "_" + i for i in obj.roles(user=user)]
        return r
